#include "..\Header\GeometryEngine.h"
#include "..\Header\MshFile.h"
#include "..\Header\OglViewerWidget.h"
#include "..\Header\MainWindow.h"
#include "..\Header\tga.h"
#include <cmath>
#include <QRegExp>


/////////////////////////////////////////////////////////////////////////
// public constructor/destructor

GeometryEngine::GeometryEngine(QObject *parent)
	: QObject(parent)
	, m_indexBuf(QOpenGLBuffer::IndexBuffer)
{
	initializeOpenGLFunctions();
}

GeometryEngine::~GeometryEngine()
{
	clearData();
}


/////////////////////////////////////////////////////////////////////////
// private functions

void GeometryEngine::loadTexture(QString filePath, QString fileName)
{
	bool loadSuccess(false);
	QImage img;

	if (fileName.isEmpty())
	{
		loadSuccess = true;
		img = QImage(1, 1, QImage::Format_RGB32);
		img.fill(Qt::red);
	}
	else
		img = loadTga(filePath + "/" + fileName, loadSuccess);

	if (!loadSuccess)
		emit sendMessage("WARNING: texture not found or corrupted: " + QString(fileName), 1);
	
	// Load image to OglTexture
	QOpenGLTexture* new_texture = new QOpenGLTexture(img.mirrored());

	// Set nearest filtering mode for texture minification
	new_texture->setMinificationFilter(QOpenGLTexture::Nearest);

	// Set bilinear filtering mode for texture magnification
	new_texture->setMagnificationFilter(QOpenGLTexture::Linear);

	// Wrap texture coordinates by repeating
	// f.ex. texture coordinate (1.1, 1.2) is same as (0.1, 0.2)
	new_texture->setWrapMode(QOpenGLTexture::Repeat);

	m_textures.push_back(new_texture);
}

void GeometryEngine::clearData()
{
	if (m_arrayBuf.isCreated())
		m_arrayBuf.destroy();
	if (m_indexBuf.isCreated())
		m_indexBuf.destroy();

	for (auto it : m_textures)
		delete it;
	m_textures.clear();
	m_drawList.clear();
}


/////////////////////////////////////////////////////////////////////////
// public slots

void GeometryEngine::loadFile(QString filePath)
{
	// cleanup old stuff and recreate buffers
	clearData();
	m_arrayBuf.create();
	m_indexBuf.create();

	//reset view
	emit requestResetView();
	emit sendMessage("loading file..", 0);

	try
	{
		QVector<Model*>* models;
		QStringList* textureNames;
		QVector<VertexData> vertexData;
		QVector<GLuint> indexData;

		// open file and get the information
		MshFile file(filePath, this);

		models = file.getModels();
		textureNames = file.getTextureNames();
		m_boundings = file.getBoundingBox();

		// collect data
		unsigned int indexOffset(0);
		unsigned int vertexOffset(0);
		for (auto& modelIterator : *models)
		{
			for (auto& segmentIterator : modelIterator->segmList)
			{
				// get draw information
				DrawInformation new_info;
				new_info.offset = indexOffset;
				new_info.size = segmentIterator->indices.size();
				new_info.textureIndex = segmentIterator->textureIndex;
				new_info.modelMatrix = modelIterator->m4x4Translation;
				new_info.modelMatrix.rotate(modelIterator->quadRotation);

				// add offset to indices, no need to do it for the first one (maybe it's very big)
				if (vertexOffset != 0)
					for (auto& it : segmentIterator->indices)
						it += vertexOffset;

				// save data
				vertexData += segmentIterator->vertices;
				indexData += segmentIterator->indices;
				m_drawList.push_back(new_info);

				// update offset
				indexOffset += new_info.size;
				vertexOffset += segmentIterator->vertices.size();
			}
		}

		// Transfer vertex data to VBO 0
		m_arrayBuf.bind();
		m_arrayBuf.allocate(vertexData.data(), vertexData.size() * sizeof(VertexData));

		// Transfer index data to VBO 1
		m_indexBuf.bind();
		m_indexBuf.allocate(indexData.data(), indexData.size() * sizeof(GLuint));

		emit sendMessage("loading textures..", 0);

		// load the textures
		int split = filePath.lastIndexOf(QRegExp("/|\\\\"));
		for (auto& it : *textureNames)
			loadTexture(filePath.left(split), it);

		loadTexture("", "");

		emit requestUpdate();
		emit sendMessage("done..", 0);
		emit sendFileInfo(filePath.right(filePath.size() - split - 1), QStringList(*textureNames), vertexData.size(), indexData.size() / 3);
	}
	catch (std::invalid_argument e)
	{
		clearData();
		emit sendMessage(QString(e.what()), 2);
	}
}

void GeometryEngine::drawGeometry(QOpenGLShaderProgram *program)
{
	if (!m_arrayBuf.isCreated() || !m_indexBuf.isCreated())
		return;

// Setup
	// Tell OpenGL which VBOs to use
	m_arrayBuf.bind();
	m_indexBuf.bind();

	// Allways normalize by this
	QMatrix4x4 normMatrix;
	float maxExtent = std::max(std::max(m_boundings.extents[0], m_boundings.extents[1]), m_boundings.extents[2]);
	normMatrix.scale(1 / maxExtent);
	normMatrix.translate(-m_boundings.center[0], -m_boundings.center[1], -m_boundings.center[2]);
	program->setUniformValue("norm_matrix", normMatrix);

	// Allways use texture unit 0
	program->setUniformValue("texture", 0);

	// Offset for position
	quintptr offset = 0;

	// Tell OpenGL programmable pipeline how to locate vertex position data
	int vertexLocation = program->attributeLocation("a_position");
	program->enableAttributeArray(vertexLocation);
	program->setAttributeBuffer(vertexLocation, GL_FLOAT, offset, 3, sizeof(VertexData));

	// Offset for texture coordinate
	offset += sizeof(QVector3D);

	// Tell OpenGL programmable pipeline how to locate vertex texture coordinate data
	int texcoordLocation = program->attributeLocation("a_texcoord");
	program->enableAttributeArray(texcoordLocation);
	program->setAttributeBuffer(texcoordLocation, GL_FLOAT, offset, 2, sizeof(VertexData));

	//Offset for normal
	offset += sizeof(QVector2D);

	// Tell OpenGL programmable pipeline how to locate vertex normal data
	int normLocation = program->attributeLocation("a_normal");
	program->enableAttributeArray(normLocation);
	program->setAttributeBuffer(normLocation, GL_FLOAT, offset, 3, sizeof(VertexData));

// Paint

	for (auto& it : m_drawList)
	{
		// bind the correct texture
		if (it.textureIndex < m_textures.size())
			m_textures.at(it.textureIndex)->bind();
		else
			m_textures.last()->bind();

		// Set model matrix
		program->setUniformValue("m_matrix", it.modelMatrix);

		// Draw cube geometry using indices from VBO 1
		glDrawElements(GL_TRIANGLES, it.size, GL_UNSIGNED_INT, (void*)(it.offset * sizeof(GLuint)));
	}
}

